#!/usr/bin/env python3

import argparse
import cherrypy
import configparser
import csv
import mimetypes
import os
import pathlib
import statsd
from io import StringIO
from cherrypy import log as log
from cherrypy._cpdispatch import Dispatcher
from datetime import datetime

# CherryPy uses mimetypes module to determine the content-type.
mimetypes.types_map['.csv'] = 'text/csv'

class Index(object):

    def __init__(self, config):
        """
        Process requests for /
        """
        self.config = config

    def index(*self, **kwargs):
        """
        Sometimes an empty GET is done by some watchdogs to see if we are
        accessible.
        """
        print("INDEX:")
        return


class DataBus(object):

    def __init__(self, config):
        """
        Process requests for Data drive D:
        """
        self.config = config

    def process_request(self, *args, **kwargs):
        print ("CherryPY: DATABUS:  process_request", args, kwargs)
        return


class DataDrive(object):

    def __init__(self, config):
        """
        Process requests for Data drive D:
        """
        self.config = config
        self.metric = statsd.StatsClient('localhost', 8125, prefix='fsserver')

    def index(*self, **kwargs):
        """
        Sometimes an empty GET is done by some watchdogs to see if we are
        accessible.
        """
        return

    def process_data_path_req(self, *args, **kwargs):
        """
        Process requests of path http://<url>/data
        """
        print ("CherryPY: process_data_path_req", args, kwargs)

        # pdb.set_trace()
        # example /D/Data
        path = kwargs['path_info']
        posix_path = self._get_local_path('data/' + path.lower())

        if not posix_path:
            raise cherrypy.HTTPError(
                status=404,
                message="FSSERVER: Path could not be processed:"
                + posix_path.name)
            return

        return self._process_request(posix_path)

    def process_drive_path_req(self, *args, **kwargs):
        """
        Process requests of path http://<url>/drived/
        """
        print ("CherryPY: process_drive_path_req", args, kwargs)

        # pdb.set_trace()
        # example /D/Data
        path = kwargs['path_info']
        posix_path = self._get_local_path(path.lower())

        if not posix_path:
            raise cherrypy.HTTPError(
                status=404,
                message="FSSERVER: Path could not be processed:"
                + posix_path.name)
            return

        return self._process_request(posix_path)

    def process_winpath_req(self, *args, **kwargs):
        print ("CherryPY: process_winpath_req", args, kwargs)

        win_path = kwargs['win_file_path']

        # Remove the leading '\'
        if win_path[0] == '\\':
            wp = pathlib.PureWindowsPath(win_path[1:])
        else:
            wp = pathlib.PureWindowsPath(win_path)

        posix_path = self._get_local_path(wp.as_posix().lower())
        if (not posix_path.name):
            raise cherrypy.HTTPError(
                status=404,
                message="FSSERVER: Path could not be processed:"
                + posix_path.name)
            return

        return self._process_request(posix_path)

    def _process_request(self, posix_path):
        if not posix_path.exists():
            raise cherrypy.HTTPError(
                status=404,
                message="FSSERVER: File/Dir not found :" + posix_path.name)
            return

        if posix_path.is_file():
            self.metric.incr('request.file')
            return self._process_file_request(posix_path)

        if posix_path.is_dir():
            self.metric.incr('request.dir')
            return self._process_dir_request(posix_path)

    def _set_file_headers(self, file_path):
        self._set_common_headers()

        file_stat = file_path.stat()

        cherrypy.response.headers['X-FileSize'] = file_stat.st_size
        cherrypy.response.headers['FSServer-Content-Length'] = file_stat.st_size
        #cherrypy.response.headers['Created'] = '01D0999AE8B2B318'
        #cherrypy.response.headers['Last-Modified'] = '01D0999AE8B3284B'

        # Convert to windows file time hex string. This is how AP
        # needs it
        win_file_ctime = self._posixtime_to_win_filetime(file_stat.st_ctime)
        cherrypy.response.headers['Created'] = win_file_ctime

        win_file_mtime = self._posixtime_to_win_filetime(file_stat.st_mtime)
        cherrypy.response.headers['Last-Modified'] = win_file_mtime
        cherrypy.response.headers['Modified'] = win_file_mtime

        return

    def _posixtime_to_win_filetime(self, posix_time_secs):
        # Windows FILETIME = 100 nanoseconds intervals since Jan 1, 1601, 64-bit.
        # Posix time = 100 second intervals sine Jan 1, 1970
        # Nano Seconds since Jan 1, 1601 to Jan 1, 1970
        #pdb.set_trace()

        epoch_diff_ns = 116444736000000000
        mtime_ns = posix_time_secs * 10000000
        win_file_time_ns = int(mtime_ns) + epoch_diff_ns
        #high_32 = hex(win_file_time_ns >> 32)[2:].upper()
        #low_32 = hex(win_file_time_ns & 0x00000000ffffffff)[2:].upper()
        #file_time_str = ''.join((((8 - len(high_32)) * '0'), high_32, low_32))

        hex_str = hex(win_file_time_ns)[2:].upper()

        # Prefix the hex string with 0 to fill 16 bytes
        file_time_str = ''.join((((16 - len(hex_str)) * '0'), hex_str))

        return file_time_str

    def _set_common_headers(self):
        # Dont send Server=CherryPy
        #del cherrypy.response.headers['Server']

        # fsserver Windows doesn't send Content-Type, Server headers
        #del cherrypy.response.headers['Content-Type']

        # Connection is not persistent
        cherrypy.response.headers['Connection'] = 'Close'

    def _get_local_path(self, path):
        try:
            #import pdb; pdb.set_trace()
            app_root = self.config['LINUX']['AppDir']
            data_root = self.config['LINUX']['DataDir']
            # 'data/servicemanager/servicestatus.csv' becomes
            # '/usr/local/ap/data/servicemanager/servicestatus.csv'
            if path.startswith('data'):
                posix_path = pathlib.PosixPath(data_root).joinpath(
                    pathlib.PurePosixPath(path).relative_to('data'))
            elif path.startswith('app'):
                posix_path = pathlib.PosixPath(app_root).joinpath(
                    pathlib.PurePosixPath(path).relative_to('app'))
            else:
                posix_path = pathlib.PosixPath('')

        except ValueError as e:
            print("_get_local_path - ValueError :", e)

        finally:
            return posix_path

    def _process_file_request(self, posix_path):
        if (os.access(str(posix_path), os.R_OK)) is False:
            raise cherrypy.HTTPError(
                status=403,
                message="FSSERVER: No permission to read :" + str(posix_path))
            return

        return self._sendfile(posix_path)

    def _time_to_ap_readable(self, timestamp):
        dt = datetime.fromtimestamp(timestamp)
        return dt.strftime('%Y-%m-%d %H:%M:%S')

    def _process_dir_request(self, posix_path):
        # Build the response in a in-memory csv
        in_mem_text = StringIO()
        in_mem_text.write('#Fields:')

        fields = ['filename', 'size', 'attributes', 'creationtime',
                  'lastaccesstime', 'lastwritetime']
        writer = csv.DictWriter(in_mem_text, fieldnames=fields)
        writer.writeheader()

        curr_dir_stat = posix_path.stat()
        writer.writerow({
            'filename' : '.',
            'size' : curr_dir_stat.st_size,
            'attributes':'D',
            'creationtime' : self._time_to_ap_readable(curr_dir_stat.st_ctime),
            'lastaccesstime' : self._time_to_ap_readable(curr_dir_stat.st_atime),
            'lastwritetime' : self._time_to_ap_readable(curr_dir_stat.st_mtime)
        })

        parent_dir_stat = posix_path.parent.stat()
        writer.writerow({
            'filename':'..',
            'size':parent_dir_stat.st_size,
            'attributes':'D',
            'creationtime' : self._time_to_ap_readable(parent_dir_stat.st_ctime),
            'lastaccesstime' : self._time_to_ap_readable(parent_dir_stat.st_atime),
            'lastwritetime' : self._time_to_ap_readable(parent_dir_stat.st_mtime)
        })

        for file in posix_path.iterdir():
            fstat = file.stat()
            if file.is_dir() :
                attribute = 'D'
            else:
                attribute = 'A'

            writer.writerow({
                'filename': file.name,
                'size': fstat.st_size,
                'attributes': attribute,
                'creationtime': self._time_to_ap_readable(fstat.st_ctime),
                'lastaccesstime': self._time_to_ap_readable(fstat.st_atime),
                'lastwritetime': self._time_to_ap_readable(fstat.st_mtime)
            })
        in_mem_text.write('#EOF')

        return self._send_dir_listing(in_mem_text)

    def _send_dir_listing(self, stream):
        self._set_common_headers()

        # Go to the beginning of the stream
        stream.seek(0)

        return stream

    def _sendfile(self, file_path):
        self._set_file_headers(file_path)

        # with opens and closes the file_path
        # Need to do buffered io and return
        #cherrypy.lib.static.serve_file(file_path)
        #if "servicestatus" in file_path:
        try:
            f = file_path.open('rb')

            # TODO: Does cherry py close the file?. How to defer close in
            # python?. Maybe the GC closes after cherrpy releases the file
            # handle
            return f
            #with open(file_path) as f:
                #return f.read()
        except EOFError as e:
            self.metric.incr('file.error.404')
            raise cherrypy.HTTPError(
                status=404,
                message="FSSERVER: Empty File, could not be read:"
                + file_path.name + e)

        except IOError as e:
            self.metric.incr('file.error.403')
            raise cherrypy.HTTPError(
                status=403,
                message="FSSERVER: File exists, but could not be read:"
                + file_path.name + e)


def setup_request_dispatcher(config):
    data_drive_controller = DataDrive(config)

    # cherrypy internally imports routes package
    dispatcher = cherrypy.dispatch.RoutesDispatcher()

    # Handling DataDrive file requests. Order of registration Matters.
    drive_d_alias = ['drived', 'driveD', 'DriveD', 'DRIVED']
    for alias in drive_d_alias:
        dispatcher.connect(name='', route='/'+ alias + '/{path_info:.*}',
                           controller=data_drive_controller, action="process_drive_path_req")
        dispatcher.connect(name='', route='/' + alias + '{win_file_path}',
                           controller=data_drive_controller, action="process_winpath_req")

    # TODO: We might need to enable /data/ too.
    #dispatcher.connect(name='', route='/data/{path_info:.*}',
                       #controller=data_drive_controller, action="process_data_path_req")

    drive_d_alias = ['d', 'D']
    for alias in drive_d_alias:
        dispatcher.connect(name='', route='/'+ alias + '/{path_info:.*}',
                           controller=data_drive_controller, action="process_drive_path_req")
        dispatcher.connect(name='', route='/' + alias + '{win_file_path}',
                           controller=data_drive_controller, action="process_winpath_req")

   #dispatcher.connect(name='', route='/data{win_file_path}',
                       #controller=data_drive_controller, action="process_winpath_req")
    """

    dispatcher.connect(name='', route='/d/{path_info:.*}',
                       controller=data_drive_controller, action="process_drive_path_req")
    dispatcher.connect(name='', route='/D/{path_info:.*}',
                       controller=data_drive_controller, action="process_drive_path_req")
    dispatcher.connect(name='', route='/driveD/{path_info:.*}',
                       controller=data_drive_controller, action="process_drive_path_req")
    dispatcher.connect(name='', route='/drived/{path_info:.*}',
                       controller=data_drive_controller, action="process_drive_path_req")
    dispatcher.connect(name='', route='/DriveD/{path_info:.*}',
                       controller=data_drive_controller, action="process_drive_path_req")
    dispatcher.connect(name='', route='/DRIVED/{path_info:.*}',
                       controller=data_drive_controller, action="process_drive_path_req")

    dispatcher.connect(name='', route='/drived{win_file_path}',
                       controller=data_drive_controller, action="process_winpath_req")
    dispatcher.connect(name='', route='/driveD{win_file_path}',
                       controller=data_drive_controller, action="process_winpath_req")
    dispatcher.connect(name='', route='/DriveD{win_file_path}',
                       controller=data_drive_controller, action="process_winpath_req")
    dispatcher.connect(name='', route='/d{win_file_path}',
                       controller=data_drive_controller, action="process_winpath_req")
    dispatcher.connect(name='', route='/D{win_file_path}',
                       controller=data_drive_controller, action="process_winpath_req")
        """

    # Return empty result for ServiceManager Commands
    dispatcher.connect(name='', route='/ServiceManager/{path_info:.*}',
                       controller=data_drive_controller, action="index")
    #dispatcher.connect(name='', route='/DataBus',
                       #controller=data_bus, action="process_request")

    # TODO: have a handler for incorrect paths sent
    #dispatcher.connect(name='', route='/', controller=root, action="index")
    dispatcher.connect(name='', route='/', controller=data_drive_controller, action="index")

    return dispatcher


class force_lower_dispatcher(Dispatcher):
    """
    Callers call using different cases. Make all lower case and let
    cherrypy dispatch.
    """
    def __call__(self, path):
        return Dispatcher.__call__(self, path.lower())


def load_fsserver_ini(filename):
    # TODO: Validate Config before loading
    config = configparser.ConfigParser()
    config.read(filename)
    return config


def setup_ssl_listener(config):
    # Setup the SSL listener
    server = cherrypy._cpserver.Server()
    server.socket_host = '0.0.0.0'
    server.socket_port = int(config['SSLHosting']['Port'])
    server.ssl_module = 'builtin'

    cert_dir = config['Certs']['Directory']
    cert = '/'.join((cert_dir, config['Certs']['SSLCert']))
    priv_key = '/'.join((cert_dir, config['Certs']['SSLPrivateKey']))

    if (os.access(cert, os.R_OK)) is False:
        log.error('SSL Cert not found :' + cert)
        quit()

    if (os.access(priv_key, os.R_OK)) is False:
        log.error('SSL Private Key not found :' + priv_key)
        quit()

    server.ssl_certificate = cert
    server.ssl_private_key = priv_key

    server.subscribe()


def setup_http_listener(config):
    # Setup the HTTP listener
    server = cherrypy._cpserver.Server()
    server.socket_host = '0.0.0.0'
    server.socket_port = int(config['HTTPHosting']['Port'])
    server.subscribe()
    return server

def setup_cherrypy(fs_config):
    dispatcher = setup_request_dispatcher(fs_config)

    cp_config = {
        '/': {
            'request.dispatch': dispatcher,
        }
    }

    # Setup cherrypy
    cherrypy.tree.mount(root=None, config=cp_config)
    cherrypy.server.unsubscribe()

    #setup_ssl_listener(fs_config)
    setup_http_listener(fs_config)

    #try:
    cherrypy.engine.signals.subscribe()

    return cherrypy

def start_and_block(cherrypy):
    cherrypy.engine.start()
    cherrypy.engine.block()

    #except SSLError as e:
        #print ("SSL ERROR CAUGHT : " + e)

def main(config_file):
    # TODO: Ensure another instance is not running
    fs_config = load_fsserver_ini(config_file)

    cherrypy = setup_cherrypy(fs_config)

    start_and_block(cherrypy)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Start AP FSServer")
    parser.add_argument('config_file', metavar='CONFIG.INI', help='Config File ini')

    args = parser.parse_args()
    config_file = args.config_file
    #if not os.path.exists('/usr/local/ap/app/fsserver/fsserver.ini'):
    if not os.path.exists(config_file):
        print("ERROR: fsserver.ini could not be found")
        from sys import exit
        exit(1)

    #main('/usr/local/ap/app/fsserver/fsserver.ini')
    main(config_file)
