#!/usr/bin/env python3

import argparse
import fsserver
import unittest
import requests
import tempfile
import shutil
from configparser import ConfigParser
from random import randint
from pathlib import PosixPath
#from os import getcwd, path, makedirs


class FSServerE2ETest(unittest.TestCase):
    """
    Start FSServer on a random port and run a set of tests.
    """
    cherrypy = ''
    http_port = 6698

    @classmethod
    def setUpClass(cls):
        test_fsserver_ini = get_fsserver_ini()

        # Get a random port
        cls.http_port = str(randint(13100, 13200))

        test_fsserver_ini['HTTPHosting']['Port'] = cls.http_port

        cls.data_dir = tempfile.mkdtemp(suffix='_ap_unittest_data')
        test_fsserver_ini['LINUX']['DataDir'] = cls.data_dir

        cls.app_dir = tempfile.mkdtemp(suffix='_ap_unittest_app')
        test_fsserver_ini['LINUX']['AppDir']  = cls.app_dir

        #test_fsserver_ini['LINUX']['AppDir'] = getcwd()
        cls.cherrypy = fsserver.setup_cherrypy(test_fsserver_ini)

        cls.cherrypy.engine.start()

    @classmethod
    def tearDownClass(cls):
        cls.cherrypy.engine.exit()
        shutil.rmtree(cls.data_dir, ignore_errors=True)
        shutil.rmtree(cls.app_dir, ignore_errors=True)

    def test_data_file_read(self):
        # TEST_THIS_COMMENT_EXISTS
        url = 'http://localhost:' + self.http_port + r'/d\data'

    def test_dir_list_child_dir(self):
        pass

    def test_dir_listing_win_path(self):
        for drive_alias in ['d', 'drived', 'DriveD', 'driveD', 'DRIVED']:
            for path_root in ['\\app', '\\data']:
                url = 'http://localhost:' + self.http_port + r'/' + drive_alias + path_root
                with self.subTest(url=url):

                    # Create a tmp file and check for the same in dir listing
                    if path_root == '\\app':
                        with tempfile.NamedTemporaryFile(dir=self.app_dir) as tmpfile:
                            resp = requests.get(url, timeout=10)
                            file_name = PosixPath(tmpfile.name).name
                            verify_dir_listing(self, resp, file_name)
                    else:
                        with tempfile.NamedTemporaryFile(dir=self.data_dir) as tmpfile:
                            resp = requests.get(url, timeout=10)
                            file_name = PosixPath(tmpfile.name).name
                            verify_dir_listing(self, resp, file_name)


    def test_dir_listing_posix_path(self):
        for drive_alias in ['d', 'drived', 'DriveD', 'driveD', 'DRIVED']:
            for path_root in ['/app', '/data']:
                #url = 'http://localhost:' + self.http_port + r'/' + drive_alias + '/data'
                url = 'http://localhost:' + self.http_port + r'/' + drive_alias + path_root
                with self.subTest(url=url):

                    # Create a tmp file and check for the same in dir listing
                    if path_root == '/app':
                        with tempfile.NamedTemporaryFile(dir=self.app_dir) as tmpfile:
                            resp = requests.get(url, timeout=10)
                            file_name = PosixPath(tmpfile.name).name
                            #import pdb;pdb.set_trace()
                            verify_dir_listing(self, resp, file_name)
                    else:
                        with tempfile.NamedTemporaryFile(dir=self.data_dir) as tmpfile:
                            resp = requests.get(url, timeout=10)
                            file_name = PosixPath(tmpfile.name).name
                            #import pdb;pdb.set_trace()
                            verify_dir_listing(self, resp, file_name)


    def test_non_existent_dir_listing(self):
        """
        Test a non-existent dir listing returns 404
        """
        url = 'http://localhost:' + self.http_port + r'/d\data\NON_EXISTENT'
        resp = requests.get(url, timeout=10)
        self.assertEqual(resp.status_code, 404,
                         "Expected HTTP Status Code 404")

def verify_dir_listing(self, resp, verify_file):
    self.assertEqual(resp.status_code, 200,
                     "Expected HTTP Status Code 200")

    lines = resp.text.splitlines()
    self.assertTrue(test_fsserver_headers(lines[0]),
                    "Did not return the correct dir listing header.")

    self.assertTrue(lines[1].startswith(r'.,4096,D'),
                    "Does not contain current dir info at second line.")

    self.assertTrue(lines[2].startswith(r'..,4096,D'),
                    "Does not contain parent dir info at third line.")

    file_found = False

    #import pdb;pdb.set_trace()
    for line in lines:
        if line.startswith(verify_file):
            file_found = True
            break

    #import pdb; pdb.set_trace()
    self.assertTrue(file_found,
                    verify_file + ' not found in dir listing.')
    print ("RESPONSE :\n", resp.text)


def test_fsserver_headers(content):
    first_line = r'#Fields:filename,size,attributes,creationtime,lastaccesstime,lastwritetime'
    return content.startswith(first_line)


def get_fsserver_ini():
    """
    Return a fsserver.ini config object
    """
    config = ConfigParser()
    config.read_string("""
            [CherryPy]
            URLPathCaseSensitive=off

            [Certs]
            Directory=./certs
            SSLCert=CCPSCH010261732.crt
            SSLPrivateKey=CCPSCH010261732.key

            [SSLHosting]
            Port=6798

            [HTTPHosting]
            Port=6698

            [LINUX]
            AppDir=/tmp/ap/app
            DataDir=/tmp/ap/data
                      """)
    return config


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-d')

    args = parser.parse_args()
    print (args)
    unittest.main()
